# coding=utf-8
"""
    港股每一手交易数量模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os

import pandas as pd

from ..CoreBu.ABuBase import FreezeAttrMixin
from ..CoreBu import ABuEnv
from ..CoreBu.ABuFixes import six
from ..UtilBu.ABuDTUtil import singleton
from ..MarketBu.ABuSymbol import Symbol

__author__ = '阿布'
__weixin__ = 'abu_quant'

_rom_dir = ABuEnv.g_project_rom_data_dir
"""文件定期重新爬取，更新"""
_hk_unit_csv = os.path.join(_rom_dir, 'hk_unit.csv')
"""默认每一手股数1000"""
K_DEFAULT_UNIT = 1000


@singleton
class AbuHkUnit(FreezeAttrMixin):
    """AbuHkUnit单例，混入FreezeAttrMixin在__init__中冻结了接口，外部只可以读取"""

    def __init__(self):
        """
            self.hk_unit_df表结构如下所示：只有一个列unit代表每一手股数，行代表港股symbol
                    unit
            hk02011	2000
            hk01396	2000
            hk08112	4800
            hk08198	4000
            hk01143	4000
            ............
        """
        # 读取本地csv到内存，由于AbuHkUnit单例只进行一次
        self.hk_unit_df = pd.read_csv(_hk_unit_csv, index_col=0)
        # __init__中使用FreezeAttrMixin._freeze冻结了接口
        self._freeze()

    def query_unit(self, symbol):
        """
        对外查询接口，查询对应symbol每一手交易数量
        :param symbol: 可以是Symbol对象，也可以是symbol字符串对象
        """
        if isinstance(symbol, Symbol):
            # Symbol对象进行转换
            symbol = symbol.value
        elif isinstance(symbol, six.string_types) and symbol.isdigit():
            # symbol字符串, 但是没有hk，则加上
            symbol = 'hk{}'.format(symbol)

        # noinspection PyBroadException
        try:
            unit = self.hk_unit_df.loc[symbol].values[0]
        except:
            # 查询失败赋予默认值
            unit = K_DEFAULT_UNIT
        return unit

    def __str__(self):
        """打印对象显示：hk_unit_df.info， hk_unit_df.describe"""
        return 'info:\n{}\ndescribe:\n{}'.format(self.hk_unit_df.info(),
                                                 self.hk_unit_df.describe())

    __repr__ = __str__

    def __len__(self):
        """对象长度：hk_unit_df.shape[0]，即行数"""
        return self.hk_unit_df.shape[0]

    def __contains__(self, item):
        """
        成员测试：标准化item后检测item是否在self.hk_unit_df.index中
        :param item: 可以是Symbol对象，也可以是symbol字符串对象
        """
        if isinstance(item, Symbol):
            item = item.value
        elif isinstance(item, six.string_types) and item.isdigit():
            item = 'hk{}'.format(item)

        return item in self.hk_unit_df.index

    def __getitem__(self, key):
        """索引获取：套接self.query_unit(key)"""
        return self.query_unit(key)

    def __setitem__(self, key, value):
        """索引设置：对外抛出错误， 即不准许外部设置"""
        raise AttributeError("AbuHkUnit set value!!!")


"""由于封装对外所以不使用模块单例"""
# """模块单例"""
# single_hk_unit = AbuHkUnit()



# -*- encoding:utf-8 -*-
"""
    选股并行多任务调度模块
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import itertools
import logging


from .ABuPickStockExecute import do_pick_stock_work, do_pick_stock_thread_work
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketDataFetchMode
from ..CoreBu.ABuEnvProcess import AbuEnvProcess
from ..MarketBu.ABuMarket import split_k_market, all_symbol
from ..MarketBu import ABuMarket
from ..CoreBu.ABuFixes import partial
from ..CoreBu.ABuParallel import delayed, Parallel
from ..CoreBu.ABuDeprecated import AbuDeprecated


class AbuPickStockMaster(object):
    """选股并行多任务调度类"""

    @classmethod
    def do_pick_stock_with_process(cls, capital, benchmark, stock_pickers, choice_symbols=None,
                                   n_process_pick_stock=ABuEnv.g_cpu_cnt,
                                   callback=None):
        """
        选股并行多任务对外接口
        :param capital: 资金类AbuCapital实例化对象
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param stock_pickers: 选股因子序列
        :param choice_symbols: 初始备选交易对象序列
        :param n_process_pick_stock: 控制启动多少进程并行选股操作
        :param callback: 并行选股工作函数
        :return: 最终选股结果序列
        """
        input_choice_symbols = True
        if choice_symbols is None or len(choice_symbols) == 0:
            choice_symbols = all_symbol()
            input_choice_symbols = False

        if n_process_pick_stock <= 0:
            # 因为下面要n_process > 1做判断而且要根据n_process_pick_stock来split_k_market
            n_process_pick_stock = ABuEnv.g_cpu_cnt
        if stock_pickers is not None:

            # TODO 需要区分hdf5和csv不同存贮情况，csv存贮模式下可以并行读写
            # 只有E_DATA_FETCH_FORCE_LOCAL才进行多任务模式，否则回滚到单进程模式n_process = 1
            if n_process_pick_stock > 1 and ABuEnv.g_data_fetch_mode != EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL:
                # 1. hdf5多进程容易写坏数据，所以只多进程读取，不并行写入
                # 2. MAC OS 10.9 之后并行联网＋numpy 系统bug crash，卡死等问题
                logging.info('batch get only support E_DATA_FETCH_FORCE_LOCAL for Parallel!')
                n_process_pick_stock = 1

            # 根据输入的choice_symbols和要并行的进程数，分配symbol到n_process_pick_stock进程中
            process_symbols = split_k_market(n_process_pick_stock, market_symbols=choice_symbols)

            # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
            if n_process_pick_stock > 1:
                n_process_pick_stock = len(process_symbols)

            parallel = Parallel(
                n_jobs=n_process_pick_stock, verbose=0, pre_dispatch='2*n_jobs')

            if callback is None:
                callback = do_pick_stock_work

            # do_pick_stock_work被装饰器add_process_env_sig装饰，需要进程间内存拷贝对象AbuEnvProcess
            p_nev = AbuEnvProcess()
            # 开始并行任务执行
            out_choice_symbols = parallel(delayed(callback)(choice_symbols, benchmark,
                                                            capital,
                                                            stock_pickers, env=p_nev)
                                          for choice_symbols in process_symbols)

            # 将每一个进程返回的选股序列合并成一个序列
            choice_symbols = list(itertools.chain.from_iterable(out_choice_symbols))

        """
            如下通过env中的设置来切割训练集，测试集数据，或者直接使用训练集，测试集，
            注意现在的设置有优先级，即g_enable_last_split_test > g_enable_last_split_train > g_enable_train_test_split
            TODO: 使用enum替换g_enable_last_split_test， g_enable_last_split_train， g_enable_train_test_split设置
        """
        if not input_choice_symbols and ABuEnv.g_enable_last_split_test:
            # 只使用上次切割好的测试集交易对象
            choice_symbols = ABuMarket.market_last_split_test()
        elif not input_choice_symbols and ABuEnv.g_enable_last_split_train:
            # 只使用上次切割好的训练集交易对象
            choice_symbols = ABuMarket.market_last_split_train()
        elif ABuEnv.g_enable_train_test_split:
            # 切割训练集交易对象与测试集交易对象，返回训练集交易对象
            choice_symbols = ABuMarket.market_train_test_split(ABuEnv.g_split_tt_n_folds, choice_symbols)

        return choice_symbols

    @classmethod
    @AbuDeprecated('hdf5 store mode will crash or dead!')
    def do_pick_stock_with_process_mix_thread(cls, capital, benchmark, stock_pickers, choice_symbols=None, n_process=8,
                                              n_thread=10):
        """Deprecated不应该使用，因为默认hdf5多线程读取会有问题"""
        callback = partial(do_pick_stock_thread_work, n_thread=n_thread)
        return cls.do_pick_stock_with_process(capital, benchmark, stock_pickers, choice_symbols=choice_symbols,
                                              n_process_pick_stock=n_process, callback=callback)



# -*- encoding:utf-8 -*-
"""选股因子参数以及选择图形可视化"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import abstractmethod

import ipywidgets as widgets

from ..WidgetBu.ABuWGBase import WidgetFactorBase, WidgetFactorManagerBase
from ..WidgetBu.ABuWGBFBase import BFSubscriberMixin

__author__ = '阿布'
__weixin__ = 'abu_quant'


class PickStockWGManager(WidgetFactorManagerBase):
    """选股策略组织类"""

    def _init_widget(self):
        """构建内置的卖出策略可视化组件，构造出self.factor_box"""

        from ..WidgetBu.ABuWGPickStock import PSPriceWidget, PSRegressAngWidget
        from ..WidgetBu.ABuWGPickStock import PSShiftDistanceWidget, PSNTopWidget
        self.ps_array = []
        self.ps_array.append(PSPriceWidget(self))
        self.ps_array.append(PSRegressAngWidget(self))
        self.ps_array.append(PSShiftDistanceWidget(self))
        self.ps_array.append(PSNTopWidget(self))

        #  ps() call用widget组list
        children = [ps() for ps in self.ps_array]
        if self.scroll_factor_box:
            self.factor_box = widgets.Box(children=children,
                                          layout=self.factor_layout)
        else:
            # 一行显示两个，2个为一组，组装sub_children_group序列,
            sub_children_group = self._sub_children(children, len(children) / 2)
            sub_children_box = [widgets.HBox(sub_children) for sub_children in sub_children_group]
            self.factor_box = widgets.VBox(sub_children_box)
        # 买入因子是特殊的存在，都需要买入因子的全局数据
        self.buy_factor_manger = None

    def seed_choice_symbol_update(self, seed_choice_symbol):
        """更新股池中的种子symbol到需要种子symbol的策略中"""
        if seed_choice_symbol is None or len(seed_choice_symbol) == 0:
            return

        for ps in self.ps_array:
            if hasattr(ps, 'seed_choice_symbol_key'):
                # 策略中要更新种子symbol的需要定义函数, 定义一个key进行更新
                cs_key = ps.seed_choice_symbol_key()

                for factor in self.factor_dict.values():
                    # 从全局选股策略中寻找这个策略
                    if factor['class'] == ps.delegate_class():
                        # 如果找到将种子symbol进行更新
                        factor.update({cs_key: seed_choice_symbol})

                for factor in self.buy_factor_manger.factor_dict.values():
                    # 从附属于买入因子的选股策略中进行搜索更新
                    if 'stock_pickers' in factor:
                        for ps_factor in factor['stock_pickers']:
                            if ps_factor['class'] == ps.delegate_class():
                                ps_factor.update({cs_key: seed_choice_symbol})
                                # from ..UtilBu.ABuOsUtil import show_msg
                                # show_msg(cs_key, str(list(factor.keys())))

    def register(self, buy_factor_manger):
        """选股manager内部因子注册接收买入因子添加的改变"""
        self.buy_factor_manger = buy_factor_manger
        for ps in self.ps_array:
            self.buy_factor_manger.register_subscriber(ps)


class WidgetPickStockBase(WidgetFactorBase, BFSubscriberMixin):
    """选股策略可视化基础类"""

    def _pick_stock_base_ui(self):
        """选股策略中通用ui: xd, reversed初始构建"""
        xd_tip = widgets.Label(u'设置选股策略生效周期，默认252天',
                               layout=widgets.Layout(width='300px', align_items='stretch'))
        self.xd = widgets.IntSlider(
            value=252,
            min=1,
            max=252,
            step=1,
            description=u'周期',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.xd_box = widgets.VBox([xd_tip, self.xd])

        reversed_tip = widgets.Label(u'反转选股结果，默认不反转',
                                     layout=widgets.Layout(width='300px', align_items='stretch'))
        self.reversed = widgets.Checkbox(
            value=False,
            description=u'反转结果',
            disabled=False,
        )
        self.reversed_box = widgets.VBox([reversed_tip, self.reversed])

    def __init__(self, wg_manager):
        super(WidgetPickStockBase, self).__init__(wg_manager)
        self.add = widgets.Button(description=u'添加为全局选股策略', layout=widgets.Layout(width='98%'),
                                  button_style='info')
        # 添加全局选股策略指令按钮
        self.add.on_click(self.add_pick_stock)
        # 运行混入的BFSubscriberMixin中ui初始化
        self.subscriber_ui([u'点击\'已添加的买入策略\'框中的买入策略', u'将选股策略做为买入策略的附属选股策略'])
        # 买入策略框点击行为：将本卖出策略加到对应的买入策略做为附属
        self.buy_factors.observe(self.add_pick_stock_to_buy_factor, names='value')
        self.accordion.set_title(0, u'添加为指定买入因子的选股策略')
        self.accordion.selected_index = -1
        self.add_box = widgets.VBox([self.add, self.accordion])
        # 构建选股策略独有基础通用ui
        self._pick_stock_base_ui()
        # 具体子策略构建
        self._init_widget()

    @abstractmethod
    def _init_widget(self):
        """子类因子界面设置初始化"""
        pass

    @abstractmethod
    def make_pick_stock_unique(self):
        """
            子类因子构建唯一描述以及因子字典
            返回值两个：
            1. 因子构建字典对象
            2. 因子唯一描述
        """
        pass

    # noinspection PyUnusedLocal
    def add_pick_stock(self, bt):
        """对应按钮添加全局策略"""
        # 构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key
        factor_dict, factor_desc_key = self.make_pick_stock_unique()
        self.wg_manager.add_factor(factor_dict, factor_desc_key)

    def add_pick_stock_to_buy_factor(self, select):
        """对应按钮添加策略到指定买入策略中"""
        self.add_to_buy_factor(select, self.make_pick_stock_unique, 'stock_pickers')



# coding=utf-8
"""
    内置数据源示例实现模块：

    所有数据接口仅供学习使用，以及最基本使用测试，如需进一步使用，请购买数据
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os

import random
import math
import sqlite3 as sqlite

import pandas as pd

from ..CoreBu.ABuEnv import EMarketTargetType, EMarketSubType
from ..CoreBu import ABuEnv
from ..MarketBu import ABuNetWork
from ..MarketBu.ABuDataBase import StockBaseMarket, SupportMixin, FuturesBaseMarket, TCBaseMarket
from ..MarketBu.ABuDataParser import BDParser, TXParser, NTParser, SNUSParser
from ..MarketBu.ABuDataParser import SNFuturesParser, SNFuturesGBParser, HBTCParser
from ..UtilBu import ABuStrUtil, ABuDateUtil, ABuMd5
from ..UtilBu.ABuDTUtil import catch_error
from ..CoreBu.ABuDeprecated import AbuDeprecated
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange, range, filter

"""网络请求（连接10秒，接收60秒）超时时间"""
K_TIME_OUT = (10, 60)


def random_from_list(array):
    """从参数array中随机取一个元素"""
    # 在array长度短的情况下，测试比np.random.choice效率要高
    return array[random.randrange(0, len(array))]


@AbuDeprecated('only read old symbol db, miss update!!!')
def query_symbol_sub_market(symbol):
    path = TXApi.K_SYMBOLS_DB
    conn = sqlite.connect(path)
    cur = conn.cursor()
    symbol = symbol.lower()
    query = "select {} from {} where {} like \'{}.%\'".format(TXApi.K_DB_TABLE_SN, TXApi.K_DB_TABLE_NAME,
                                                              TXApi.K_DB_TABLE_SN, symbol)
    cur.execute(query)
    results = cur.fetchall()
    conn.close()
    sub_market = ''
    if results is not None and len(results) > 0:
        try:
            if results[0][0].find('.') > 0:
                sub_market = '.' + results[0][0].split('.')[1].upper()
        except:
            logging.info(results)
    return sub_market


@catch_error(return_val=None, log=False)
def query_symbol_from_pinyin(pinyin):
    """通过拼音对symbol进行模糊查询"""
    path = TXApi.K_SYMBOLS_DB
    conn = sqlite.connect(path)
    cur = conn.cursor()
    pinyin = pinyin.lower()
    query = "select stockCode from {} where pinyin=\'{}\'".format(TXApi.K_DB_TABLE_NAME, pinyin)
    cur.execute(query)
    results = cur.fetchall()
    conn.close()
    if len(results) > 0:
        code = results[0][0]
        # 查询到的stcok code eg：sh111111，usabcd.n
        start = 2
        end = len(code)
        if '.' in code:
            # 如果是美股要截取.
            end = code.find('.')
        return code[start:end]


class BDApi(StockBaseMarket, SupportMixin):
    """bd数据源，支持港股，美股，a股"""

    K_NET_CONNECT_START = '&start='
    K_NET_DAY = 'http://gp.baidu.com:80/stocks/stockkline?from=android&os_ver=21&format=json&vv=3.3.0' \
                '&uid=&BDUSS=&cuid=%s&channel=default_channel&device=%s&logid=%s&actionid=%s&device_net_type' \
                '=wifi&period=day&stock_code=%s&fq_type=front'

    MINUTE_NET_5D = 'http://gp.baidu.com:80/stocks/stocktimelinefive?from=android&os_ver=21&format=json' \
                    '&vv=3.3&uid=&BDUSS=&cuid=%s&channel=default_channel&device=%s&logid=%s&actionid=%s' \
                    '&device_net_type=wifi&stock_code=%s&step=10'

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        super(BDApi, self).__init__(symbol)
        self._action_id = int(ABuDateUtil.time_seconds())
        self._version2_log_cnt = 0
        self.data_parser_cls = BDParser

    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        self._version2_log_cnt += 1
        log_id = self._action_id + self._version2_log_cnt * 66
        cuid = ABuStrUtil.create_random_with_num_low(40)
        device = random_from_list(StockBaseMarket.K_DEV_MODE_LIST)
        url = BDApi.K_NET_DAY % (cuid, device, str(log_id), str(self._action_id), self._symbol.value)
        # logging.info(url)
        next_start = None
        kl_df = None
        if start:
            # 需重新计算n_fold
            days = ABuDateUtil.diff(start, ABuDateUtil.current_str_date(), check_order=False)
            # 每次返回300条数据
            n_folds = int(days / 300.0)

        for _ in xrange(0, n_folds):
            if next_start:
                url = url + BDApi.K_NET_CONNECT_START + str(next_start)
            # logging.info(url)
            data = ABuNetWork.get(url=url, timeout=K_TIME_OUT)
            temp_df = None
            if data is not None:
                temp_df = self.data_parser_cls(self._symbol, data.json()).df

            if temp_df is not None:
                next_start = int(temp_df.loc[temp_df.index[0], ['date']].values[0])
            kl_df = temp_df if kl_df is None else pd.concat([temp_df, kl_df])
            # 因为是从前向后请求，且与时间无关，所以可以直接在for里面中断
            if kl_df is None:
                return None

            """由于每次放回300条>1年的数据，所以超出总数就不再请求下一组"""
            if kl_df.shape[0] > ABuEnv.g_market_trade_year * n_folds:
                break

        return StockBaseMarket._fix_kline_pd(kl_df, n_folds, start, end)

    def minute(self, n_folds=5, *args, **kwargs):
        self._version2_log_cnt += 1
        cuid = ABuStrUtil.create_random_with_num_low(40)
        log_id = self._action_id + self._version2_log_cnt * 66
        device = random_from_list(StockBaseMarket.K_DEV_MODE_LIST)
        url = BDApi.MINUTE_NET_5D % (cuid, device, str(log_id), str(self._action_id), self._symbol.value)

        return ABuNetWork.get(url=url, timeout=K_TIME_OUT).json()


class TXApi(StockBaseMarket, SupportMixin):
    """tx数据源，支持港股，美股，a股"""

    K_NET_BASE = "http://ifzq.gtimg.cn/appstock/app/%sfqkline/get?p=1&param=%s,day,,,%d," \
                 "qfq&_appName=android&_dev=%s&_devId=%s&_mid=%s&_md5mid=%s&_appver=4.2.2&_ifChId=303&_screenW=%d" \
                 "&_screenH=%d&_osVer=%s&_uin=10000&_wxuin=20000&__random_suffix=%d"

    K_NET_HK_MNY = 'http://proxy.finance.qq.com/ifzqgtimg/stock/corp/hkmoney/sumary?' \
                   'symbol=%s&type=sum&jianjie=1&_appName=android' \
                   '&_dev=%s&_devId=%s&_mid=%s&_md5mid=%s&_appver=5.5.0&_ifChId=277' \
                   '&_screenW=%d&_screenH=%d&_osVer=%s&_uin=10000&_wxuin=20000&_net=WIFI&__random_suffix=%d'

    K_DB_TABLE_NAME = "values_table"
    K_DB_TABLE_SN = "stockCode"
    p_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.path.pardir))
    K_SYMBOLS_DB = os.path.join(p_dir, 'RomDataBu/symbols_db.db')

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        super(TXApi, self).__init__(symbol)
        # 设置数据源解析对象类
        self.data_parser_cls = TXParser

    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        cuid = ABuStrUtil.create_random_with_num_low(40)
        cuid_md5 = ABuMd5.md5_from_binary(cuid)
        random_suffix = ABuStrUtil.create_random_with_num(5)
        dev_mod = random_from_list(StockBaseMarket.K_DEV_MODE_LIST)
        os_ver = random_from_list(StockBaseMarket.K_OS_VERSION_LIST)
        screen = random_from_list(StockBaseMarket.K_PHONE_SCREEN)

        days = ABuEnv.g_market_trade_year * n_folds + 1
        # start 不为空时计算 获取天数，获取的数据肯定比预期的数据多，因为同一时间内，交易日的天数一定不比实际的天数多
        if start:
            temp_end = ABuDateUtil.current_str_date()
            days = ABuDateUtil.diff(start, temp_end, check_order=False)

        sub_market = None
        if self._symbol.market == EMarketTargetType.E_MARKET_TARGET_US:
            # sub_market = self.query_symbol_sub_market(self._symbol.value)
            market = self._symbol.market.value
            if '.' in self._symbol.value:
                # 如果已经有.了说明是大盘，大盘不需要子市场，eg：us.IXIC
                sub_market = ''
            else:
                # 这里tx的source不支持US_PINK, US_OTC, US_PREIPO
                sub_market_map = {EMarketSubType.US_N.value: 'n',

                                  EMarketSubType.US_PINK.value: 'n',
                                  EMarketSubType.US_OTC.value: 'n',
                                  EMarketSubType.US_PREIPO.value: 'n',
                                  EMarketSubType.US_AMEX.value: 'n',

                                  EMarketSubType.US_OQ.value: 'oq'}
                sub_market = '.{}'.format(sub_market_map[self._symbol.sub_market.value])
            url = TXApi.K_NET_BASE % (
                market, self._symbol.value + sub_market, days,
                dev_mod, cuid, cuid, cuid_md5, screen[0], screen[1], os_ver, int(random_suffix, 10))
        elif self._symbol.market == EMarketTargetType.E_MARKET_TARGET_HK:
            market = self._symbol.market.value
            url = TXApi.K_NET_BASE % (
                market, self._symbol.value, days,
                dev_mod, cuid, cuid, cuid_md5, screen[0], screen[1], os_ver, int(random_suffix, 10))
        else:
            market = ''
            url = TXApi.K_NET_BASE % (
                market, self._symbol.value, days,
                dev_mod, cuid, cuid, cuid_md5, screen[0], screen[1], os_ver, int(random_suffix, 10))

        data = ABuNetWork.get(url, timeout=K_TIME_OUT)
        if data is not None:
            kl_pd = self.data_parser_cls(self._symbol, sub_market, data.json()).df
        else:
            return None

        return StockBaseMarket._fix_kline_pd(kl_pd, n_folds, start, end)

    def hkmoney(self):
        """港股概要信息接口"""
        if self._symbol.market != EMarketTargetType.E_MARKET_TARGET_HK:
            raise TypeError('hkmoney only support hk!!')

        cuid = ABuStrUtil.create_random_with_num_low(40)
        cuid_md5 = ABuMd5.md5_from_binary(cuid)
        random_suffix = ABuStrUtil.create_random_with_num(5)
        dev_mod = random_from_list(StockBaseMarket.K_DEV_MODE_LIST)
        os_ver = random_from_list(StockBaseMarket.K_OS_VERSION_LIST)
        screen = random_from_list(StockBaseMarket.K_PHONE_SCREEN)

        url = TXApi.K_NET_HK_MNY % (self._symbol.value, dev_mod, cuid, cuid, cuid_md5, screen[0], screen[1], os_ver,
                                    int(random_suffix, 10))
        return ABuNetWork.get(url, timeout=K_TIME_OUT)

    def minute(self, n_fold=5, *args, **kwargs):
        """分钟k线接口"""
        raise NotImplementedError('TXApi minute NotImplementedError!')


class NTApi(StockBaseMarket, SupportMixin):
    """nt数据源，支持港股，美股，a股"""

    K_NET_BASE = "http://img1.money.126.net/data/%s/kline/day/history/%d/%s.json"

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        super(NTApi, self).__init__(symbol)
        # 设置数据源解析对象类
        self.data_parser_cls = NTParser

    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        kl_df = None
        if start is None or end is None:
            end_year = int(ABuDateUtil.current_str_date()[:4])
            start_year = end_year - n_folds + 1
        else:
            start_year = int(start[:4])
            end_year = int(end[:4])
        req_year = list(range(start_year, end_year + 1))

        if self._symbol.market == EMarketTargetType.E_MARKET_TARGET_US:
            market = self._symbol.market.value
            symbol = self._symbol.symbol_code.upper()
            if self._symbol.is_us_index():
                # ntes 需要做映射匹配大盘symbol
                index_dict = {'.DJI': 'DOWJONES', '.IXIC': 'NASDAQ', '.INX': 'SP500'}
                symbol = index_dict[symbol]
        elif self._symbol.market == EMarketTargetType.E_MARKET_TARGET_HK:
            market = self._symbol.market.value
            symbol = self._symbol.symbol_code.upper()
        elif self._symbol.market == EMarketTargetType.E_MARKET_TARGET_CN:
            market = self._symbol.market.value
            symbol = self._symbol.symbol_code
            if self._symbol.is_sz_stock():
                symbol = '1{}'.format(symbol)
            else:
                symbol = '0{}'.format(symbol)
        else:
            raise TypeError('NTApi dt support {}'.format(self._symbol.market))

        for year in req_year:
            url = NTApi.K_NET_BASE % (market, year, symbol)
            data = ABuNetWork.get(url=url, retry=1, timeout=K_TIME_OUT)
            temp_df = None
            if data is not None:
                temp_df = self.data_parser_cls(self._symbol, data.json()).df
            if temp_df is not None:
                kl_df = temp_df if kl_df is None else kl_df.append(temp_df)
        if kl_df is None:
            return None
        return StockBaseMarket._fix_kline_pd(kl_df, n_folds, start, end)

    def minute(self, n_fold=5, *args, **kwargs):
        """分钟k线接口"""
        raise NotImplementedError('NTApi minute NotImplementedError!')


class SNUSApi(StockBaseMarket, SupportMixin):
    """snus数据源，支持美股"""
    K_NET_BASE = "http://stock.finance.sina.com.cn/usstock/api/json_v2.php/US_MinKService.getDailyK?" \
                 "symbol=%s&___qn=3n"

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        super(SNUSApi, self).__init__(symbol)
        # 设置数据源解析对象类
        self.data_parser_cls = SNUSParser

    def _support_market(self):
        """声明数据源支持美股"""
        return [EMarketTargetType.E_MARKET_TARGET_US]

    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        url = SNUSApi.K_NET_BASE % self._symbol.symbol_code
        data = ABuNetWork.get(url=url, timeout=K_TIME_OUT).json()
        kl_df = self.data_parser_cls(self._symbol, data).df
        if kl_df is None:
            return None
        return StockBaseMarket._fix_kline_pd(kl_df, n_folds, start, end)

    def minute(self, n_fold=5, *args, **kwargs):
        """分钟k线接口"""
        raise NotImplementedError('SNUSApi minute NotImplementedError!')


class SNFuturesApi(FuturesBaseMarket, SupportMixin):
    """sn futures数据源，支持国内期货"""

    K_NET_BASE = "http://stock.finance.sina.com.cn/futures/api/json_v2.php/" \
                 "IndexService.getInnerFuturesDailyKLine?symbol=%s"

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        super(SNFuturesApi, self).__init__(symbol)
        # 设置数据源解析对象类
        self.data_parser_cls = SNFuturesParser

    def _support_market(self):
        """声明数据源支持期货数据"""
        return [EMarketTargetType.E_MARKET_TARGET_FUTURES_CN]

    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        url = SNFuturesApi.K_NET_BASE % self._symbol.symbol_code
        data = ABuNetWork.get(url=url, timeout=K_TIME_OUT).json()
        kl_df = self.data_parser_cls(self._symbol, data).df
        if kl_df is None:
            return None
        return FuturesBaseMarket._fix_kline_pd(kl_df, n_folds, start, end)


class SNFuturesGBApi(FuturesBaseMarket, SupportMixin):
    """sn futures数据源，支持国际期货"""

    K_NET_BASE = "http://stock2.finance.sina.com.cn/futures/api/jsonp.php/" \
                 "var %s%s=/GlobalFuturesService.getGlobalFuturesDailyKLine?symbol=%s&_=%s"

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        super(SNFuturesGBApi, self).__init__(symbol)
        # 设置数据源解析对象类
        self.data_parser_cls = SNFuturesGBParser

    def _support_market(self):
        """声明数据源支持期货数据, 支持国际期货市场"""
        return [EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL]

    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        today = ABuDateUtil.current_str_date().replace('-', '_')
        url = SNFuturesGBApi.K_NET_BASE % (self._symbol.symbol_code, today, self._symbol.symbol_code, today)
        data = ABuNetWork.get(url=url, timeout=(10, 60))
        text = data.text
        # 返回的是Javascript字符串解析出dict
        js_dict = ABuNetWork.parse_js(text[text.find('=(') + 2:text.rfind(')')])
        kl_df = self.data_parser_cls(self._symbol, js_dict).df
        if kl_df is None:
            return None
        return FuturesBaseMarket._fix_kline_pd(kl_df, n_folds, start, end)


class HBApi(TCBaseMarket, SupportMixin):
    """hb数据源，支持币类，比特币，莱特币"""

    K_NET_BASE = 'https://www.huobi.com/qt/staticmarket/%s_kline_100_json.js?length=%d'

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        super(HBApi, self).__init__(symbol)
        # 设置数据源解析对象类
        self.data_parser_cls = HBTCParser

    def _support_market(self):
        """只支持币类市场"""
        return [EMarketTargetType.E_MARKET_TARGET_TC]

    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        req_cnt = n_folds * ABuEnv.g_market_trade_year
        if start is not None and end is not None:
            # 向上取整数，下面使用_fix_kline_pd再次进行剪裁, 要使用current_str_date不能是end
            folds = math.ceil(ABuDateUtil.diff(ABuDateUtil.date_str_to_int(start),
                                               ABuDateUtil.current_str_date()) / 365)
            req_cnt = folds * ABuEnv.g_market_trade_year

        url = HBApi.K_NET_BASE % (self._symbol.symbol_code, req_cnt)
        data = ABuNetWork.get(url=url, timeout=K_TIME_OUT).json()
        kl_df = self.data_parser_cls(self._symbol, data).df
        if kl_df is None:
            return None
        return TCBaseMarket._fix_kline_pd(kl_df, n_folds, start, end)

    def minute(self, *args, **kwargs):
        """分钟k线接口"""
        raise NotImplementedError('HBApi minute NotImplementedError!')



# -*- encoding:utf-8 -*-
"""
    借鉴sklearn GridSearch，针对买入因子，卖出因子，选股因子最合进行最优寻找分析
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import operator
from collections import Mapping
from functools import reduce
from ..CoreBu.ABuFixes import partial
from itertools import product

import logging
import numpy as np

from ..TradeBu.ABuBenchmark import AbuBenchmark
from ..TradeBu.ABuCapital import AbuCapital
from ..TradeBu.ABuKLManager import AbuKLManager
from .ABuMetricsScore import AbuScoreTuple, WrsmScorer, make_scorer
from ..AlphaBu.ABuPickStockMaster import AbuPickStockMaster
from ..AlphaBu.ABuPickTimeMaster import AbuPickTimeMaster
from ..CoreBu.ABuEnvProcess import add_process_env_sig, AbuEnvProcess
from ..CoreBu.ABuParallel import delayed, Parallel
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketDataFetchMode
from ..UtilBu import ABuProgress

__author__ = '阿布'
__weixin__ = 'abu_quant'


class ParameterGrid(object):
    """参数进行product辅助生成类"""

    def __init__(self, param_grid):
        if isinstance(param_grid, Mapping):
            param_grid = [param_grid]
        self.param_grid = param_grid

    def __iter__(self):
        """迭代参数组合实现"""
        for p in self.param_grid:
            items = sorted(p.items())
            if not items:
                yield {}
            else:
                keys, values = zip(*items)
                for v in product(*values):
                    params = dict(zip(keys, v))
                    yield params

    def __len__(self):
        """参数组合长度实现"""
        product_mul = partial(reduce, operator.mul)
        return sum(product_mul(len(v) for v in p.values()) if p else 1
                   for p in self.param_grid)

    def __getitem__(self, ind):
        """通过index方式获取某个参数组合实现"""
        for sub_grid in self.param_grid:
            if not sub_grid:
                if ind == 0:
                    return {}
                else:
                    ind -= 1
                    continue
            keys, values_lists = zip(*sorted(sub_grid.items())[::-1])
            sizes = [len(v_list) for v_list in values_lists]
            total = np.product(sizes)

            if ind >= total:
                ind -= total
            else:
                out = {}
                for key, v_list, n in zip(keys, values_lists, sizes):
                    ind, offset = divmod(ind, n)
                    out[key] = v_list[offset]
                return out

        raise IndexError('ParameterGrid index out of range')


def _check_param_grid(param_grid):
    """检测迭代序列是否可进行grid"""

    if hasattr(param_grid, 'items'):
        param_grid = [param_grid]

    for p in param_grid:
        for v in p.values():
            if isinstance(v, np.ndarray) and v.ndim > 1:
                raise ValueError("Parameter array should be one-dimensional.")

            check = [isinstance(v, k) for k in (list, tuple, np.ndarray)]
            if True not in check:
                raise ValueError("Parameter values should be a list.")

            if len(v) == 0:
                raise ValueError("Parameter values should be a non-empty "
                                 "list.")


@add_process_env_sig
def grid_mul_func(read_cash, benchmark, buy_factors, sell_factors, stock_pickers, choice_symbols, kl_pd_manager=None):
    """
    针对输入的买入因子，卖出因子，选股因子及其它参数，进行两年历史交易回测，返回结果包装AbuScoreTuple对象
    :param read_cash: 初始化资金数(int)
    :param benchmark: 交易基准对象，AbuBenchmark实例对象
    :param buy_factors: 买入因子序列
    :param sell_factors: 卖出因子序列
    :param stock_pickers: 选股因子序列
    :param choice_symbols: 初始备选交易对象序列
    :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
    :return: AbuScoreTuple对象
    """

    # 通过初始化资金数，交易基准对象构造资金管理对象capital
    capital = AbuCapital(read_cash, benchmark)

    # 由于grid_mul_func以处于多任务运行环境，所以不内部不再启动多任务，使用1个进程选股
    n_process_pick_stock = 1
    # 由于grid_mul_func以处于多任务运行环境，所以不内部不再启动多任务，使用1个进程择时
    n_process_pick_time = 1
    # 由于grid_mul_func以处于多任务运行环境，所以不内部不再启动多任务，使用1个进程数据收集
    n_process_kl = 1

    if stock_pickers is not None:
        # 有选股因子序列首选进行选股
        choice_symbols = AbuPickStockMaster.do_pick_stock_with_process(capital, benchmark,
                                                                       stock_pickers,
                                                                       choice_symbols=choice_symbols,
                                                                       n_process_pick_stock=n_process_pick_stock)

    if choice_symbols is None or len(choice_symbols) == 0:
        logging.info('pick stock result is zero!')
        return None

    # 通过买入因子，卖出因子等进行择时操作
    orders_pd, action_pd, all_fit_symbols_cnt = AbuPickTimeMaster.do_symbols_with_same_factors_process(
        choice_symbols, benchmark,
        buy_factors, sell_factors, capital, kl_pd_manager=kl_pd_manager, n_process_kl=n_process_kl,
        n_process_pick_time=n_process_pick_time)

    # 将最终结果包装为AbuScoreTuple对象
    return AbuScoreTuple(orders_pd, action_pd, capital, benchmark, buy_factors, sell_factors,
                         stock_pickers)


# noinspection PyAttributeOutsideInit
class GridSearch(object):
    """最优grid search对外接口类"""

    def __init__(self, read_cash, choice_symbols, stock_pickers_product=None,
                 buy_factors_product=None, sell_factors_product=None, score_weights=None, metrics_class=None):
        """
        :param read_cash: 初始化资金数(int)
        :param choice_symbols: 初始备选交易对象序列
        :param stock_pickers_product: 选股因子product之后的序列
        :param buy_factors_product: 买入因子product之后的序列
        :param sell_factors_product: 卖出因子product之后的序列
        :param score_weights: make_scorer中设置的评分权重
        :param metrics_class: make_scorer中设置的度量类
        """
        self.read_cash = read_cash
        self.benchmark = AbuBenchmark()
        self.kl_pd_manager = AbuKLManager(self.benchmark, AbuCapital(self.read_cash, self.benchmark))
        self.choice_symbols = choice_symbols
        self.stock_pickers_product = [None] if stock_pickers_product is None else stock_pickers_product
        self.buy_factors_product = [None] if buy_factors_product is None else buy_factors_product
        self.sell_factors_product = [None] if sell_factors_product is None else sell_factors_product
        self.score_weights = score_weights
        self.metrics_class = metrics_class

    def fit(self, score_class=WrsmScorer, n_jobs=-1):
        """
        开始寻找最优因子参数组合，费时操作，迭代所有因子组合进行交易回测，回测结果进行评分
        :param score_class: 对回测结果进行评分的评分类，AbuBaseScorer类型，非对象，只传递类信息
        :param n_jobs: 默认回测并行的任务数，默认-1, 即启动与cpu数量相同的进程数
        :return: (scores: 评分结果dict， score_tuple_array: 因子组合序列)
        """

        if len(self.stock_pickers_product) == 1 and self.stock_pickers_product[0] is None:
            # 如果没有设置选股因子，外层统一进行交易数据收集，之所以是1，以为在__init__中[None]的设置
            need_batch_gen = self.kl_pd_manager.filter_pick_time_choice_symbols(self.choice_symbols)
            self.kl_pd_manager.batch_get_pick_time_kl_pd(need_batch_gen, n_process=ABuEnv.g_cpu_cnt)

        # 只有E_DATA_FETCH_FORCE_LOCAL才进行多任务模式，否则回滚到单进程模式n_jobs = 1
        if n_jobs != 1 and ABuEnv.g_data_fetch_mode != EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL:
            # 1. hdf5多进程还容易写坏数据
            # 2. MAC OS 10.9 之后并行联网＋numpy 系统bug crash，卡死等问题
            logging.info('batch get only support E_DATA_FETCH_FORCE_LOCAL for Parallel!')
            n_jobs = 1

        parallel = Parallel(
            n_jobs=n_jobs, verbose=0, pre_dispatch='2*n_jobs')

        pass_kl_pd_manager = None
        if len(self.stock_pickers_product) == 1 and self.stock_pickers_product[0] is None:
            pass_kl_pd_manager = self.kl_pd_manager

        # 暂时关闭多进程进度条，太多, 注意这种全局设置一定要在AbuEnvProcess初始化之前完成
        ABuProgress.g_show_ui_progress = False
        # 多任务环境下的内存环境拷贝对象AbuEnvProcess
        p_nev = AbuEnvProcess()
        # 多层迭代各种类型因子，没一种因子组合作为参数启动一个新进程，运行grid_mul_func
        out_abu_score_tuple = parallel(
            delayed(grid_mul_func)(self.read_cash, self.benchmark, buy_factors, sell_factors, stock_pickers,
                                   self.choice_symbols, pass_kl_pd_manager, env=p_nev)
            for stock_pickers in self.stock_pickers_product for buy_factors in self.buy_factors_product for
            sell_factors in self.sell_factors_product)
        ABuProgress.g_show_ui_progress = True

        # 返回的AbuScoreTuple序列转换score_tuple_array
        score_tuple_array = list(out_abu_score_tuple)
        # 使用ABuMetricsScore中make_scorer对多个参数组合的交易结果进行评分，详情阅读ABuMetricsScore模块
        scores = make_scorer(score_tuple_array, score_class, weights=self.score_weights,
                             metrics_class=self.metrics_class)
        # 评分结果最好的赋予best_score_tuple_grid
        self.best_score_tuple_grid = score_tuple_array[scores.index[-1]]
        return scores, score_tuple_array



# -*- encoding:utf-8 -*-
"""
    函数对象的工具类
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from collections import OrderedDict
from abc import ABCMeta

from ..CoreBu.ABuFixes import six

try:
    from inspect import getfullargspec as getargspec
except ImportError:

    from inspect import getargspec

__author__ = '夜猫'
__weixin__ = 'abu_quant'


class ArgNoDefault(six.with_metaclass(ABCMeta, TypeError)):
    """没有默认值的参数对象"""
    pass


def get_func_name(func):
    """
    获取函数名称
    :param func: 传入函数
    :return: 
    """
    try:
        func_name = str(func.__name__) + '()'
    except AttributeError:
        func_name = str(func)
    return func_name


def get_arg_defaults(func):
    """
    获取函数默认值字典；没有默认值时对应NoDefaultArg对象
    :param func: 传入函数
    :return: 函数参数名：默认值
    """
    # 解包函数参数及默认值
    argspec = getargspec(func)
    spec_args = argspec.args if argspec.args else []
    defaults = argspec.defaults if argspec.defaults else ()
    # 拼装默认值dict
    no_defaults = (ArgNoDefault(),) * (len(spec_args) - len(defaults))
    args_defaults = dict(zip(spec_args, no_defaults + defaults))
    return args_defaults


def check_bind(func, *args, **kwargs):
    """
    检查要bind的对象和原函数func的参数是否对齐；对齐失败，raise TypeError
    :param func:  原函数
    :param args: 要bind的tuple对象
    :param kwargs: 要bind的dict对象
    :return: 
    """
    argspec = getargspec(func)
    spec_args = argspec.args if argspec.args else []
    # 检查 kwargs 中是否有不存在的参数名
    bad_names = set(kwargs.keys()) - set(spec_args)
    if bad_names:
        raise TypeError(
            "Got unknown arguments: {}".format(str(bad_names))
        )
    # 检查args的参数是否过长
    l_arg_len = len(args)
    if len(spec_args) < l_arg_len + len(kwargs):
        raise TypeError(
            "Function with {} arguments, but got {} arguments to bind".format(len(spec_args),
                                                                              l_arg_len + len(kwargs))
        )
    # 检查 kwargs 中是否和 args的参数冲突
    arg_inds = OrderedDict(zip(spec_args, range(len(spec_args))))
    for k, v in six.iteritems(kwargs):
        if l_arg_len > arg_inds[k]:
            raise TypeError(
                "dict argument crash on tuple argument:  {}".format(str(k))
            )


def bind_partial(func, *args, **kwargs):
    """
    绑定func的参数和对应的参数对象
    :param func: 原函数
    :param args: 要bind的tuple对象
    :param kwargs: 要bind的dict对象
    :return: 绑定后的字典
    """
    # 解包函数参数
    argspec = getargspec(func)
    spec_args = argspec.args if argspec.args else []

    # 拼装函数参数和对应的参数对象
    bind_dict = OrderedDict(zip(spec_args, [None] * len(spec_args)))
    for k, v in six.iteritems(kwargs):
        bind_dict[k] = v
    for k, v in zip(six.iterkeys(bind_dict), args):
        bind_dict[k] = v

    return bind_dict



# -*- encoding:utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import time

from ..CoreBu import ABuFixes

__author__ = '小青蛙'
__weixin__ = 'abu_quant'

BASE_XQ_URL = 'https://xueqiu.com'

BASE_XQ_HQ_URL = BASE_XQ_URL + '/hq'

BASE_XQ_STOCK_INFO = BASE_XQ_URL + '/S/'

BASE_XQ_HQ_EXCHANGE_URL = BASE_XQ_HQ_URL + '#exchange={}'

CN_INDUSTRY_LIST = 'https://xueqiu.com/industry/quote_order.json?' \
                   'size=100&order=asc&orderBy=code&exchange=CN&'
HK_INDUSTRY_LIST = 'https://xueqiu.com/stock/cata/stocklist.json?' \
                   'size=100&order=asc&orderby=code&exchange=HK&'
US_INDUSTRY_LIST = 'https://xueqiu.com/stock/cata/stocklist.json?' \
                   'size=100&order=asc&orderby=code&exchange=US&'

CN_STOCK_LIST = 'https://xueqiu.com/stock/cata/stocklist.json?' \
                'size=100&order=asc&orderby=code&type=11%2C12&'
HK_STOCK_LIST = 'https://xueqiu.com/stock/cata/stocklist.json?' \
                'size=100&order=asc&orderby=code&type=30%2C31&'
US_STOCK_LIST = 'https://xueqiu.com/stock/cata/stocklist.json?' \
                'size=100&order=asc&orderby=code&type=0%2C1%2C2%2C3&isdelay=1&'

LOGIN_URL = 'https://xueqiu.com/user/login'

# 雪球url中type值
TYPE_INFO = {
    0: 'us_stock',  # 美股(不包含中国区)
    1: 'us_a_stock_NASDAQ',  # 在拉斯达克上市的中国股票
    2: 'us_a_stock_NYSE',  # 在纽交所上市的中国股票
    3: 'us_index',  # 美国指数
    11: 'a_stock',  # A股
    12: 'a_index',  # A股指数
    30: 'hk_stock',  # 港股
    31: 'hk_index',  # 港股指数
}


class IndustryUrl(object):
    def __init__(self, market, **kwargs):
        self.market = market
        self.kwargs = kwargs
        self._base_url = None
        self.init_base_url()

    def init_base_url(self):
        if 'CN' == self.market:
            self._base_url = CN_INDUSTRY_LIST
        elif 'HK' == self.market:
            self._base_url = HK_INDUSTRY_LIST
        elif 'US' == self.market:
            self._base_url = US_INDUSTRY_LIST
        else:
            raise RuntimeError('only support {}'.format(['US', 'CN', 'HK']))

    @property
    def url(self):
        param = dict(self.kwargs)
        param['_'] = time.time()

        return self._base_url + ABuFixes.urlencode(param)


class StockListUrl(IndustryUrl):
    def init_base_url(self):
        if 'CN' == self.market:
            self._base_url = CN_STOCK_LIST
        elif 'HK' == self.market:
            self._base_url = HK_STOCK_LIST
        elif 'US' == self.market:
            self._base_url = US_STOCK_LIST
        else:
            raise RuntimeError('only support {}'.format(['US', 'CN', 'HK']))


def test():
    print(IndustryUrl('CN', plate='保险业', page=1, level2code='J68').url)
    print(IndustryUrl('US', plate='金融', page=1).url)
    print(IndustryUrl('HK', plate='保险', page=1).url)



# -*- encoding:utf-8 -*-
"""买入因子参数以及选择图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import ipywidgets as widgets

from ..WidgetBu.ABuWGBFBase import WidgetFactorBuyBase
from ..FactorBuyBu.ABuFactorBuyBreak import AbuFactorBuyBreak
from ..FactorBuyBu.ABuFactorBuyDM import AbuDoubleMaBuy
from ..FactorBuyBu.ABuFactorBuyWD import AbuFactorBuyWD
from ..FactorBuyBu.ABuFactorBuyDemo import AbuSDBreak, AbuWeekMonthBuy

__author__ = '阿布'
__weixin__ = 'abu_quant'


class BuyDMWidget(WidgetFactorBuyBase):
    """对应AbuDoubleMaBuy策略widget"""

    def _init_widget(self):
        """构建AbuDoubleMaBuy策略参数界面"""

        self.description = widgets.Textarea(
            value=u'动态自适应双均线买入策略：\n'
                  u'双均线策略是量化策略中经典的策略之一，其属于趋势跟踪策略: \n'
                  u'1. 预设两条均线：如一个ma=5，一个ma=60, 5的均线被称作快线，60的均线被称作慢线\n'
                  u'2. 择时买入策略中当快线上穿慢线（ma5上穿ma60）称为形成金叉买点信号，买入股票\n'
                  u'3. 自适应动态慢线，不需要输入慢线值，根据走势震荡套利空间，寻找合适的ma慢线\n'
                  u'4. 自适应动态快线，不需要输入快线值，根据慢线以及大盘走势，寻找合适的ma快线',
            description=u'双均线买',
            disabled=False,
            layout=self.description_layout
        )

        self.slow_label = widgets.Label(u'默认使用动态慢线，可手动固定慢线值', layout=self.label_layout)
        self.slow_int = widgets.IntSlider(
            value=60,
            min=10,
            max=120,
            step=1,
            description=u'手动',
            disabled=True,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.auto_slow = widgets.Checkbox(
            value=True,
            description=u'动态慢线',
            disabled=False
        )

        def slow_change(change):
            self.slow_int.disabled = change['new']

        self.auto_slow.observe(slow_change, names='value')
        self.slow = widgets.VBox([self.auto_slow, self.slow_int])
        self.slow_box = widgets.VBox([self.slow_label, self.slow])

        self.fast_label = widgets.Label(u'默认使用动态快线，可手动固定快线值', layout=self.label_layout)
        self.fast_int = widgets.IntSlider(
            value=5,
            min=1,
            max=90,
            step=1,
            description=u'手动',
            disabled=True,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.auto_fast = widgets.Checkbox(
            value=True,
            description=u'动态快线',
            disabled=False,
        )

        def fast_change(change):
            self.fast_int.disabled = change['new']

        self.auto_fast.observe(fast_change, names='value')
        self.fast = widgets.VBox([self.auto_fast, self.fast_int])
        self.fast_box = widgets.VBox([self.fast_label, self.fast])

        self.widget = widgets.VBox([self.description, self.slow_box, self.fast_box, self.add],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_buy_factor_unique(self):
        """对应按钮添加AbuDoubleMaBuy策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        slow_int = -1 if self.auto_slow.value else self.slow_int.value
        fast_int = -1 if self.auto_fast.value else self.fast_int.value

        factor_dict = {'class': AbuDoubleMaBuy, 'slow': slow_int, 'fast': fast_int}
        factor_desc_key = u'动态双均慢{}快{}买入'.format(
            u'动态' if slow_int == -1 else slow_int, u'动态' if fast_int == -1 else fast_int)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuDoubleMaBuy"""
        return AbuDoubleMaBuy


class BuySDWidget(WidgetFactorBuyBase):
    """对应AbuSDBreak策略widget"""

    def _init_widget(self):
        """构建AbuSDBreak策略参数界面"""

        self.description = widgets.Textarea(
            value=u'参照大盘走势向上趋势突破买入策略：\n'
                  u'在海龟突破基础上，参照大盘走势，进行降低交易频率，提高系统的稳定性处理，当大盘走势震荡时封锁交易，'
                  u'当大盘走势平稳时再次打开交易，每一个月计算一次大盘走势是否平稳',
            description=u'平稳突破',
            disabled=False,
            layout=self.description_layout
        )
        self.poly_label = widgets.Label(u'大盘走势拟合次数阀值，poly大于阀值＝震荡',
                                        layout=self.label_layout)
        self.poly = widgets.IntSlider(
            value=2,
            min=1,
            max=5,
            step=1,
            description=u'拟合',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.poly_box = widgets.VBox([self.poly_label, self.poly])
        self.xd_label = widgets.Label(u'突破周期参数：比如21，30，42天....突破',
                                      layout=self.label_layout)
        self.xd = widgets.IntSlider(
            value=21,
            min=3,
            max=120,
            step=1,
            description=u'周期',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.xd_box = widgets.VBox([self.xd_label, self.xd])
        self.widget = widgets.VBox([self.description, self.poly_box,
                                    self.xd_box, self.add],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_buy_factor_unique(self):
        """对应按钮添加AbuSDBreak策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuSDBreak, 'xd': self.xd.value, 'poly': self.poly.value}
        factor_desc_key = u'{}拟合{}天趋势突破参照大盘'.format(self.poly.value, self.xd.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuSDBreak"""
        return AbuSDBreak


class BuyWDWidget(WidgetFactorBuyBase):
    """对应AbuFactorBuyWD策略widget"""

    def _init_widget(self):
        """构建AbuFactorBuyWD策略参数界面"""

        self.description = widgets.Textarea(
            value=u'日胜率均值回复策略：\n'
                  u'1. 默认以40天为周期(8周)结合涨跌阀值计算周几适合买入\n'
                  u'2. 回测运行中每一月重新计算一次上述的周几适合买入\n'
                  u'3. 在策略日任务中买入信号为：昨天下跌，今天开盘也下跌，且明天是计算出来的上涨概率大的\'周几\'',
            description=u'周涨胜率',
            disabled=False,
            layout=self.description_layout
        )

        self.buy_dw_label = widgets.Label(u'代表周期胜率阀值，默认0.55即55%的胜率',
                                          layout=self.label_layout)
        self.buy_dw = widgets.FloatSlider(
            value=0.55,
            min=0.50,
            max=0.99,
            step=0.01,
            description=u'胜率',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='.2f',
        )
        self.buy_dw_box = widgets.VBox([self.buy_dw_label, self.buy_dw])

        self.buy_dwm_label = widgets.Label(u'代表涨幅比例阀值系数，默认0.618',
                                           layout=self.label_layout)
        self.buy_dwm = widgets.FloatSlider(
            value=0.618,
            min=0.50,
            max=1.0,
            step=0.01,
            description=u'系数',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='.3f'
        )
        self.buy_dwm_box = widgets.VBox([self.buy_dwm_label, self.buy_dwm])

        self.dw_period_label = widgets.Label(u'代表分所使用的交易周期，默认40天周期(8周)',
                                             layout=self.label_layout)
        self.dw_period = widgets.IntSlider(
            value=40,
            min=20,
            max=120,
            step=1,
            description=u'周期',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.dw_period_box = widgets.VBox([self.dw_period_label, self.dw_period])
        self.widget = widgets.VBox([self.description, self.buy_dw_box,
                                    self.buy_dwm_box, self.dw_period_box, self.add],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_buy_factor_unique(self):
        """对应按钮添加AbuFactorBuyWD策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuFactorBuyWD, 'buy_dw': self.buy_dw.value,
                       'buy_dwm': self.buy_dwm.value, 'dw_period': self.dw_period.value}
        factor_desc_key = u'日胜率{},{},{}均值回复买入'.format(
            self.buy_dw.value, self.buy_dwm.value, self.dw_period.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuFactorBuyWD"""
        return AbuFactorBuyWD


class BuyXDWidget(WidgetFactorBuyBase):
    """对应AbuFactorBuyBreak策略widget"""

    def _init_widget(self):
        """构建AbuFactorBuyBreak策略参数界面"""

        self.description = widgets.Textarea(
            value=u'海龟向上趋势突破买入策略：\n'
                  u'趋势突破定义为当天收盘价格超过N天内的最高价，超过最高价格作为买入信号买入股票持有',
            description=u'海龟买入',
            disabled=False,
            layout=self.description_layout
        )
        self.xd_label = widgets.Label(u'突破周期参数：比如21，30，42天....突破', layout=self.label_layout)
        self.xd = widgets.IntSlider(
            value=21,
            min=3,
            max=120,
            step=1,
            description=u'周期',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.xd_box = widgets.VBox([self.xd_label, self.xd])
        self.widget = widgets.VBox([self.description, self.xd_box, self.add],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_buy_factor_unique(self):
        """对应按钮添加AbuFactorBuyBreak策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuFactorBuyBreak, 'xd': self.xd.value}
        factor_desc_key = u'海龟{}天趋势突破买入'.format(self.xd.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuFactorBuyBreak"""
        return AbuFactorBuyBreak


class BuyWMWidget(WidgetFactorBuyBase):
    """对应AbuWeekMonthBuy策略widget"""

    def _init_widget(self):
        """构建AbuWeekMonthBuy策略参数界面"""

        self.description = widgets.Textarea(
            value=u'固定周期买入策略：\n'
                  u'根据参数每周买入一次或者每一个月买入一次\n'
                  u'需要与特定\'选股策略\'和\'卖出策略\'形成配合\n，'
                  u'单独使用固定周期买入策略意义不大',
            description=u'定期买入',
            disabled=False,
            layout=self.description_layout
        )

        is_buy_month_label = widgets.Label(u'可更改买入定期，默认定期一个月', layout=self.label_layout)
        self.is_buy_month = widgets.Dropdown(
            options={u'定期一个月': True, u'定期一个周': False},
            value=True,
            description=u'定期时长:',
        )
        is_buy_month_box = widgets.VBox([is_buy_month_label, self.is_buy_month])

        self.widget = widgets.VBox([self.description, is_buy_month_box, self.add],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_buy_factor_unique(self):
        """对应按钮添加AbuWeekMonthBuy策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuWeekMonthBuy, 'is_buy_month': self.is_buy_month.value}
        factor_desc_key = u'{}买入一次'.format(u'每一月' if self.is_buy_month.value else u'每一周')
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuWeekMonthBuy"""
        return AbuWeekMonthBuy



# -*- encoding:utf-8 -*-
"""
    相关系数上层技术线应用模块
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import copy
import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from statsmodels.tsa.stattools import coint

from ..CoreBu.ABuEnv import EMarketDataSplitMode
from ..CoreBu import ABuEnv
from ..MarketBu import ABuSymbolPd
from ..MarketBu.ABuSymbol import code_to_symbol
from ..SimilarBu import ABuSimilar
from ..SimilarBu import ECoreCorrType
from ..TradeBu.ABuBenchmark import AbuBenchmark
from ..UtilBu import ABuScalerUtil
from ..UtilBu.ABuProgress import do_clear_output
from ..SimilarBu.ABuSimilar import from_local
from ..TLineBu.ABuTLine import AbuTLine

__author__ = '阿布'
__weixin__ = 'abu_quant'

g_top_corr_cnt = 300
g_coint_threshold = 0.38
g_coint_show_max = 10


def rank_corr_sum(corr_df_dict, symbol):
    """
    使用corr_df_dict各个相关性矩阵中symbol的rank值进行sum合并：
        eg：rank_pd
                           pears  sperm
            usBIDU           5.0    5.0
            usFB             8.0    8.0
            usGOOG           6.0    6.0
            usNOAH           1.0    1.0
            usSFUN           7.0    7.0
            usTSLA           9.0    9.0
            usVIPS           3.0    3.0
            usWUBA           4.0    4.0

        eg: rank_pd.sum(axis=1)
            usBIDU           10.0
            usFB             16.0
            usGOOG           12.0
            usNOAH            2.0
            usSFUN           14.0
            usTSLA           18.0
            usVIPS            6.0
            usWUBA            8.0
    :param corr_df_dict: 相关性矩阵组成的字典对象
    :param symbol: eg: 'usTSLA'
    :return: rank_pd.sum(axis=1)，pd.Series对象
    """
    rank_pd = pd.DataFrame()
    for corr_df_key in corr_df_dict:
        corr_df = corr_df_dict[corr_df_key]
        """
            eg：corr_df

                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3013  0.3690  0.4015  0.3680  0.3015  0.3706  0.4320
            usFB           0.3013  1.0000  0.6609  0.2746  0.1978  0.4080  0.2856  0.2438
            usGOOG         0.3690  0.6609  1.0000  0.3682  0.1821  0.3477  0.3040  0.2917
            usNOAH         0.4015  0.2746  0.3682  1.0000  0.3628  0.2178  0.4645  0.4488
            usSFUN         0.3680  0.1978  0.1821  0.3628  1.0000  0.2513  0.2843  0.4883
            usTSLA         0.3015  0.4080  0.3477  0.2178  0.2513  1.0000  0.2327  0.3340
            usVIPS         0.3706  0.2856  0.3040  0.4645  0.2843  0.2327  1.0000  0.4189
            usWUBA         0.4320  0.2438  0.2917  0.4488  0.4883  0.3340  0.4189  1.0000
        """

        if symbol not in corr_df:
            # TODO 在这里处理有点晚
            # print(corr_df.columns)
            # print(code_to_symbol(symbol).symbol_code)
            return None

        corr_rank = corr_df[symbol].rank(ascending=False, method='first')
        """
            eg： corr_rank
            usBIDU           5.0
            usFB             8.0
            usGOOG           6.0
            usNOAH           1.0
            usSFUN           7.0
            usTSLA           9.0
            usVIPS           3.0
            usWUBA           4.0
        """
        rank_pd = rank_pd.join(pd.DataFrame(corr_rank.values, index=corr_rank.index,
                                            columns=[corr_df_key]), how='outer')
        """
            eg：rank_pd
                           pears  sperm
            usBIDU           5.0    5.0
            usFB             8.0    8.0
            usGOOG           6.0    6.0
            usNOAH           1.0    1.0
            usSFUN           7.0    7.0
            usTSLA           9.0    9.0
            usVIPS           3.0    3.0
            usWUBA           4.0    4.0
        """

    # TODO: 全是等权重的计算，需要可分配权重计算的参数
    """
        eg: rank_pd.sum(axis=1)
        usBIDU           10.0
        usFB             16.0
        usGOOG           12.0
        usNOAH            2.0
        usSFUN           14.0
        usTSLA           18.0
        usVIPS            6.0
        usWUBA            8.0
    """
    # 清一下输出，太乱
    do_clear_output()
    return rank_pd.sum(axis=1)


@from_local
def calc_similar(symbol, cmp_symbol, sum_rank=None, corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS,
                                                               ECoreCorrType.E_CORE_TYPE_SPERM), show=True):
    """
    使用ABuTLSimilar.calc_similar()函数计算返回的相关性数值是以目标股票所在市场为观察者，
    它不关心某一个股票具体相关性的数值的大小，calc_similar(a, b) 的工作流程如下：
    1.计算a与市场中所有股票的相关性
    2.将所有相关性进行rank排序
    3.查询股票b在rank序列中的位置，此位置值即为结果
    即ABuTLSimilar.calc_similar返回值由0至1，这样的好处是通过计算usTSLA与usAAPL在所有股票中的相似度水平，会更全局客观的体现相关性

    :param symbol: eg: 'usTSLA'
    :param cmp_symbol: 与symbol进行整体市场相关对比的symbol eg: 'usTSLA'
    :param sum_rank: 已经缓存了的sum_rank数据，
                        eg: sum_rank
                        usBIDU           10.0
                        usFB             16.0
                        usGOOG           12.0
                        usNOAH            2.0
                        usSFUN           14.0
                        usTSLA           18.0
                        usVIPS            6.0
                        usWUBA            8.0
    :param corr_jobs: 默认：corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS, ECoreCorrType.E_CORE_TYPE_SPERM)
                      可以再添加更多jobs
                      eg：
                        corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS, ECoreCorrType.E_CORE_TYPE_SPERM,
                                   ECoreCorrType.E_CORE_TYPE_SIGN, ECoreCorrType.E_CORE_TYPE_ROLLING)
                      注意每添加一种相关计算方法，耗时都会增加
    :param show: 是否进行可视化
    :return: rank_score (float: 0至1), sum_rank
    """

    cs_symbol = code_to_symbol(symbol)
    cs_cmp_symbol = code_to_symbol(cmp_symbol)

    if cs_symbol.market != cs_cmp_symbol.market:
        # 必须在同一个市场
        logging.info('{} and {} in different market!!!'.format(symbol, cmp_symbol))
        return

    symbol = cs_symbol.value
    cmp_symbol = cs_cmp_symbol.value
    if sum_rank is None:
        tmp_market = ABuEnv.g_market_target
        # 强制把市场设置为一样的
        ABuEnv.g_market_target = cs_symbol.market
        corr_df_dict = ABuSimilar.multi_corr_df(corr_jobs)
        # 恢复之前的市场
        ABuEnv.g_market_target = tmp_market
        """
            eg： corr_df_dict
            {'pears':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3013  0.3690  0.4015  0.3680  0.3015  0.3706  0.4320
            usFB           0.3013  1.0000  0.6609  0.2746  0.1978  0.4080  0.2856  0.2438
            usGOOG         0.3690  0.6609  1.0000  0.3682  0.1821  0.3477  0.3040  0.2917
            usNOAH         0.4015  0.2746  0.3682  1.0000  0.3628  0.2178  0.4645  0.4488
            usSFUN         0.3680  0.1978  0.1821  0.3628  1.0000  0.2513  0.2843  0.4883
            usTSLA         0.3015  0.4080  0.3477  0.2178  0.2513  1.0000  0.2327  0.3340
            usVIPS         0.3706  0.2856  0.3040  0.4645  0.2843  0.2327  1.0000  0.4189
            usWUBA         0.4320  0.2438  0.2917  0.4488  0.4883  0.3340  0.4189  1.0000

            'sperm':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3888  0.4549  0.4184  0.3747  0.3623  0.4333  0.4396
            usFB           0.3888  1.0000  0.7013  0.2927  0.2379  0.4200  0.3123  0.2216
            usGOOG         0.4549  0.7013  1.0000  0.3797  0.2413  0.3871  0.3922  0.3035
            usNOAH         0.4184  0.2927  0.3797  1.0000  0.3581  0.2066  0.4643  0.4382
            usSFUN         0.3747  0.2379  0.2413  0.3581  1.0000  0.2645  0.3890  0.4693
            usTSLA         0.3623  0.4200  0.3871  0.2066  0.2645  1.0000  0.2540  0.2801
            usVIPS         0.4333  0.3123  0.3922  0.4643  0.3890  0.2540  1.0000  0.4080
            usWUBA         0.4396  0.2216  0.3035  0.4382  0.4693  0.2801  0.4080  1.0000 }
        """
        sum_rank = rank_corr_sum(corr_df_dict, symbol)
        """
            eg: sum_rank
            usBIDU           10.0
            usFB             16.0
            usGOOG           12.0
            usNOAH            2.0
            usSFUN           14.0
            usTSLA           18.0
            usVIPS            6.0
            usWUBA            8.0
        """
        if sum_rank is None:
            logging.info('{} not in corr df!!!'.format(symbol))
            return None, None

    if cmp_symbol not in sum_rank.index:
        logging.info('{} not in sum_rank.index'.format(cmp_symbol))
        return None, None

    # sum_rank.sort_values(ascending=True)之后的结果index即是对比的排序结果值cmp_rank
    cmp_rank = sum_rank.sort_values(ascending=True).index.tolist().index(cmp_symbol)
    """
        eg: sum_rank.sort_values(ascending=True)
        usNOAH            2.0
        us_NYSE:.IXIC     4.0
        usVIPS            6.0
        usWUBA            8.0
        usBIDU           10.0
        usGOOG           12.0
        usSFUN           14.0
        usFB             16.0
        usTSLA           18.0
    """
    # 计算cmp_rank在整体sum_rank的比例位置rank_score
    rank_score = 1 - cmp_rank / sum_rank.shape[0]

    if show:
        log_func = logging.info if ABuEnv.g_is_ipython else print
        log_func('{} similar rank score {} : {}'.format(symbol, cmp_symbol, rank_score))

        # 通过make_kl_df序列的接口获取两个金融时间序列
        mul_pd = ABuSymbolPd.make_kl_df([symbol, cmp_symbol], n_folds=2)
        kl_pd = mul_pd[symbol]
        kl_pd_cmp = mul_pd[cmp_symbol]

        # 缩放到同一个数量级type_look='look_max'
        kl_pd, kl_pd_cmp = ABuScalerUtil.scaler_xy(kl_pd.close,
                                                   kl_pd_cmp.close,
                                                   type_look='look_max')
        # 首先可视化已经缩放到一个级别的两个金融序列
        kl_pd.plot()
        kl_pd_cmp.plot()
        plt.legend([symbol, cmp_symbol])
        plt.title('similar draw')
        plt.show()

        distance = (kl_pd - kl_pd_cmp)
        # 通过distance构造技术线对象AbuTLine，可视化几个技术线
        line = AbuTLine(distance, line_name='{} distance {}'.format(symbol, cmp_symbol))
        line.show()
        # 可视化技术线拟合曲线及上下拟合通道曲线
        line.show_regress_trend_channel()
        # 可视化技术线'路程位移比'
        # line.show_shift_distance()
        # 对技术线阻力位和支撑位进行绘制
        # line.show_support_resistance_trend(show=False)

    return rank_score, sum_rank


@from_local
def calc_similar_top(symbol, sum_rank=None, corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS,
                                                       ECoreCorrType.E_CORE_TYPE_SPERM), show=True, show_cnt=10):
    """
    使用corr_jobs种相关算法在env所在的市场中寻找与symbol最相关的show_cnt个，可视化
    show_cnt个
    :param symbol: eg: 'usTSLA'
    :param sum_rank: 已经缓存了的sum_rank数据，
                        eg: sum_rank
                        usBIDU           10.0
                        usFB             16.0
                        usGOOG           12.0
                        usNOAH            2.0
                        usSFUN           14.0
                        usTSLA           18.0
                        usVIPS            6.0
                        usWUBA            8.0
    :param corr_jobs: 默认：corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS, ECoreCorrType.E_CORE_TYPE_SPERM)
                      可以再添加更多jobs
                      eg：
                        corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS, ECoreCorrType.E_CORE_TYPE_SPERM,
                                   ECoreCorrType.E_CORE_TYPE_SIGN, ECoreCorrType.E_CORE_TYPE_ROLLING)
                      注意每添加一种相关计算方法，耗时都会增加
    :param show: 是否进行可视化
    :param show_cnt: 可视化最相关的top n的个数
    :return: 排序好的top 相关序列，pd.Series对象
                eg：
                    us_NYSE:.IXIC     4.0
                    usFB              6.0
                    usGOOG            8.0
                    usWUBA           11.0
                    usBIDU           11.0
                    usSFUN           14.0
                    usVIPS           16.0
                    usNOAH           18.0
    """
    cs = code_to_symbol(symbol)
    symbol = cs.value
    if sum_rank is None:
        # TODO 重复代码太多，提前头装饰器
        tmp_market = ABuEnv.g_market_target
        # 强制把市场设置为一样的
        ABuEnv.g_market_target = cs.market
        corr_df_dict = ABuSimilar.multi_corr_df(corr_jobs)
        # 恢复之前的市场
        ABuEnv.g_market_target = tmp_market
        """
            eg： corr_df_dict
            {'pears':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3013  0.3690  0.4015  0.3680  0.3015  0.3706  0.4320
            usFB           0.3013  1.0000  0.6609  0.2746  0.1978  0.4080  0.2856  0.2438
            usGOOG         0.3690  0.6609  1.0000  0.3682  0.1821  0.3477  0.3040  0.2917
            usNOAH         0.4015  0.2746  0.3682  1.0000  0.3628  0.2178  0.4645  0.4488
            usSFUN         0.3680  0.1978  0.1821  0.3628  1.0000  0.2513  0.2843  0.4883
            usTSLA         0.3015  0.4080  0.3477  0.2178  0.2513  1.0000  0.2327  0.3340
            usVIPS         0.3706  0.2856  0.3040  0.4645  0.2843  0.2327  1.0000  0.4189
            usWUBA         0.4320  0.2438  0.2917  0.4488  0.4883  0.3340  0.4189  1.0000

            'sperm':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3888  0.4549  0.4184  0.3747  0.3623  0.4333  0.4396
            usFB           0.3888  1.0000  0.7013  0.2927  0.2379  0.4200  0.3123  0.2216
            usGOOG         0.4549  0.7013  1.0000  0.3797  0.2413  0.3871  0.3922  0.3035
            usNOAH         0.4184  0.2927  0.3797  1.0000  0.3581  0.2066  0.4643  0.4382
            usSFUN         0.3747  0.2379  0.2413  0.3581  1.0000  0.2645  0.3890  0.4693
            usTSLA         0.3623  0.4200  0.3871  0.2066  0.2645  1.0000  0.2540  0.2801
            usVIPS         0.4333  0.3123  0.3922  0.4643  0.3890  0.2540  1.0000  0.4080
            usWUBA         0.4396  0.2216  0.3035  0.4382  0.4693  0.2801  0.4080  1.0000 }
        """
        sum_rank = rank_corr_sum(corr_df_dict, symbol)
        """
            eg: sum_rank
            usBIDU           10.0
            usFB             16.0
            usGOOG           12.0
            usNOAH            2.0
            usSFUN           14.0
            usTSLA           18.0
            usVIPS            6.0
            usWUBA            8.0
        """
        if sum_rank is None:
            logging.info('{} not in corr df!!!'.format(symbol))
            return

    show_cnt = sum_rank.shape[0] - 1 if show_cnt > sum_rank.shape[0] else show_cnt
    # sort_values后即为rank排序结果按照参数show_cnt个数进行返回，第一个是自身掠过
    rank_head = sum_rank.sort_values(ascending=True)[1: show_cnt + 1]
    if show:
        kl_pd = ABuSymbolPd.make_kl_df(symbol, n_folds=1)
        # 获取top show_cnt个symbol金融时间序列
        mul_pd = ABuSymbolPd.make_kl_df(rank_head.index, n_folds=1)

        for cmp_symbol in rank_head.index:
            # 迭代所有的收盘价格序列进行数据缩放
            kl_pd_close, kl_pd_cmp_close = ABuScalerUtil.scaler_xy(kl_pd.close,
                                                                   mul_pd[cmp_symbol].close,
                                                                   type_look='look_max')
            # 缩放后的数据进行可视化对比
            kl_pd_close.plot()
            kl_pd_cmp_close.plot()
            plt.legend([symbol, cmp_symbol])
            plt.title('similar draw')
            plt.show()

    """
        eg：rank_head
        us_NYSE:.IXIC     4.0
        usFB              6.0
        usGOOG            8.0
        usWUBA           11.0
        usBIDU           11.0
        usSFUN           14.0
        usVIPS           16.0
        usNOAH           18.0
    """
    return rank_head


@from_local
def coint_similar(symbol, sum_rank=None, corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS,
                                                    ECoreCorrType.E_CORE_TYPE_SPERM), show=True):
    """
    首先找到的是最相关的top个，从top n个最相关的再找协整，只考虑pvalue，因为已经是从top n个最相关的再找协整
    可视化整个过程

    :param symbol: eg: 'usTSLA'
    :param sum_rank: 已经缓存了的sum_rank数据，
                        eg: sum_rank
                        usBIDU           10.0
                        usFB             16.0
                        usGOOG           12.0
                        usNOAH            2.0
                        usSFUN           14.0
                        usTSLA           18.0
                        usVIPS            6.0
                        usWUBA            8.0
    :param corr_jobs: 默认：corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS, ECoreCorrType.E_CORE_TYPE_SPERM)
                      可以再添加更多jobs
                      eg：
                        corr_jobs=(ECoreCorrType.E_CORE_TYPE_PEARS, ECoreCorrType.E_CORE_TYPE_SPERM,
                                   ECoreCorrType.E_CORE_TYPE_SIGN, ECoreCorrType.E_CORE_TYPE_ROLLING)
                      注意每添加一种相关计算方法，耗时都会增加
    :param show: 是否进行可视化
    """
    cs = code_to_symbol(symbol)
    symbol = cs.value
    if sum_rank is None:
        tmp_market = ABuEnv.g_market_target
        # 强制把市场设置为一样的
        ABuEnv.g_market_target = cs.market
        corr_df_dict = ABuSimilar.multi_corr_df(corr_jobs)
        # 恢复之前的市场
        ABuEnv.g_market_target = tmp_market
        """
            eg： corr_df_dict
            {'pears':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3013  0.3690  0.4015  0.3680  0.3015  0.3706  0.4320
            usFB           0.3013  1.0000  0.6609  0.2746  0.1978  0.4080  0.2856  0.2438
            usGOOG         0.3690  0.6609  1.0000  0.3682  0.1821  0.3477  0.3040  0.2917
            usNOAH         0.4015  0.2746  0.3682  1.0000  0.3628  0.2178  0.4645  0.4488
            usSFUN         0.3680  0.1978  0.1821  0.3628  1.0000  0.2513  0.2843  0.4883
            usTSLA         0.3015  0.4080  0.3477  0.2178  0.2513  1.0000  0.2327  0.3340
            usVIPS         0.3706  0.2856  0.3040  0.4645  0.2843  0.2327  1.0000  0.4189
            usWUBA         0.4320  0.2438  0.2917  0.4488  0.4883  0.3340  0.4189  1.0000

            'sperm':
                            usBIDU    usFB  usGOOG  usNOAH  usSFUN  usTSLA  usVIPS  usWUBA
            usBIDU         1.0000  0.3888  0.4549  0.4184  0.3747  0.3623  0.4333  0.4396
            usFB           0.3888  1.0000  0.7013  0.2927  0.2379  0.4200  0.3123  0.2216
            usGOOG         0.4549  0.7013  1.0000  0.3797  0.2413  0.3871  0.3922  0.3035
            usNOAH         0.4184  0.2927  0.3797  1.0000  0.3581  0.2066  0.4643  0.4382
            usSFUN         0.3747  0.2379  0.2413  0.3581  1.0000  0.2645  0.3890  0.4693
            usTSLA         0.3623  0.4200  0.3871  0.2066  0.2645  1.0000  0.2540  0.2801
            usVIPS         0.4333  0.3123  0.3922  0.4643  0.3890  0.2540  1.0000  0.4080
            usWUBA         0.4396  0.2216  0.3035  0.4382  0.4693  0.2801  0.4080  1.0000 }
        """
        sum_rank = rank_corr_sum(corr_df_dict, symbol)
        """
            eg: sum_rank
            usBIDU           10.0
            usFB             16.0
            usGOOG           12.0
            usNOAH            2.0
            usSFUN           14.0
            usTSLA           18.0
            usVIPS            6.0
            usWUBA            8.0
        """
        if sum_rank is None:
            logging.info('{} not in corr df!!!'.format(symbol))
            return None, None

    top_cnt = sum_rank.shape[0] if g_top_corr_cnt > sum_rank.shape[0] else g_top_corr_cnt
    # 首先找到的是最相关的top个
    rank_head = sum_rank.sort_values(ascending=True)[1:top_cnt]

    # 使用symbol做标尺
    benchmark = AbuBenchmark(symbol, n_folds=1)
    # benchmark做为数据标尺获取最相关的top个金融时间数据
    mul_pd = ABuSymbolPd.make_kl_df(rank_head.index, n_folds=1,
                                    data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO,
                                    benchmark=benchmark)

    coint_dict = {}
    for ind, cmp_symbol in enumerate(rank_head.index):
        if cmp_symbol not in mul_pd:
            continue

        klpd_cmp = mul_pd[cmp_symbol]
        if klpd_cmp is None:
            continue

        """
                coint返回值三个如下：
                coint_t : float
                    t-statistic of unit-root test on residuals
                pvalue : float
                    MacKinnon's approximate p-value based on MacKinnon (1994)
                crit_value : dict
                    Critical values for the test statistic at the 1 %, 5 %, and 10 %
                    levels.

                这里只考虑pvalue，因为已经是从top n个最相关的再找协整
        """
        _, pvalue, _ = coint(benchmark.kl_pd.close, klpd_cmp.close)
        if pvalue < g_coint_threshold:
            # pvalue小于阀值即加入coint_dict字典

            # 记录ind为了发现取多少个sort_values(ascending=True)[1:g_top_corr_cnt]能有良好的数据
            # 即为了事后调整g_top_corr_cnt使用，并非实际需要
            coint_dict[cmp_symbol] = (pvalue, ind)
    p_value_sorted = sorted(zip(coint_dict.values(), coint_dict.keys()))
    if len(p_value_sorted) == 0:
        logging.info(
            'len(p_value_sorted) == 0 please try change tl.similar.g_top_corr_cnt|tl.similar.g_coint_threshold!')
        return None, None

    if show:
        cmp_cnt = np.minimum(len(p_value_sorted), g_coint_show_max)
        # 只取item[1]，[0]是ind
        symbols = [item[1] for item in p_value_sorted[:cmp_cnt]]
        mul_pd_swap = mul_pd.swapaxes('items', 'minor')
        close_panel_pd = mul_pd_swap['close'][symbols]
        """
            转轴后只取收盘价格序列
            eg： close_panel_pd
                          usFB  usGOOG  usNOAH  usVIPS  usWUBA  us_NYSE:.IXIC
            2015-07-24   96.95  623.56   23.40  20.250   65.25       5088.629
            2015-07-27   94.17  627.26   22.16  19.990   62.89       5039.776
            2015-07-28   95.29  628.00   22.94  20.200   60.32       5089.207
            2015-07-29   96.99  631.93   23.35  20.260   59.89       5111.730
            2015-07-30   95.21  632.59   22.87  19.700   60.24       5128.785
            ...            ...     ...     ...     ...     ...            ...
            2016-07-20  121.92  741.19   25.11  13.630   48.17       5089.930
            2016-07-21  120.61  738.63   25.51  13.690   49.25       5073.900
            2016-07-22  121.00  742.74   25.50  13.510   49.21       5100.160
            2016-07-25  121.63  739.77   25.57  13.390   49.84       5097.628
            2016-07-26  121.64  740.92   24.75  13.655   50.36       5084.629
        """
        # 将数据scale到一个级别上，注意使用mean_how=True，避免极值的干扰
        close_panel_pd = ABuScalerUtil.scaler_matrix(close_panel_pd, mean_how=True)
        """
            ABuScalerUtil.scaler_matrix缩放后的数据矩阵如下所示
            eg： close_panel_pd
                             usFB     usGOOG     usNOAH     usVIPS     usWUBA
            2015-07-24  4451.7674  4311.1198  4477.3494  6601.2284  5980.4246
            2015-07-27  4324.1148  4336.7006  4240.0882  6516.4719  5764.1211
            2015-07-28  4375.5432  4341.8168  4389.3332  6584.9290  5528.5703
            2015-07-29  4453.6041  4368.9877  4467.7825  6604.4882  5489.1591
            ...               ...        ...        ...        ...        ...
            2016-07-20  5598.3443  5124.3808  4804.5404  4443.1972  4414.9740
            2016-07-21  5538.1915  5106.6817  4881.0762  4462.7564  4513.9603
            2016-07-22  5556.0995  5135.0971  4879.1628  4404.0788  4510.2942
            2016-07-25  5585.0280  5114.5633  4892.5566  4364.9604  4568.0362
            2016-07-26  5585.4872  5122.5141  4735.6581  4451.3468  4615.6963
        """
        # 可视化scaler_matrix操作后的close
        close_panel_pd.plot()
        plt.title('close panel pd scaler_matrix')
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.show()

        close_panel_pd_cp = copy.deepcopy(close_panel_pd)

        for col in close_panel_pd_cp.columns:
            """
                做一个一摸一样的pd就是为了得到投票权重表便于运算：

                     close_panel_pd_cp[col] = benchmark.kl_pd.close

                将所有数据列都使用标尺的数据进行替换，结果是每一列的数据都相同，
                比如这样，列数据都和标尺一样
                            usFB  usGOOG  usNOAH  usVIPS  usWUBA  us_NYSE:.IXIC
                2015-07-24  265.41  265.41  265.41  265.41  265.41         265.41
                2015-07-27  253.01  253.01  253.01  253.01  253.01         253.01
                2015-07-28  264.82  264.82  264.82  264.82  264.82         264.82
                2015-07-29  263.82  263.82  263.82  263.82  263.82         263.82
                2015-07-30  266.79  266.79  266.79  266.79  266.79         266.79
                ...            ...     ...     ...     ...     ...            ...
                2016-07-20  228.36  228.36  228.36  228.36  228.36         228.36
                2016-07-21  220.50  220.50  220.50  220.50  220.50         220.50
                2016-07-22  222.27  222.27  222.27  222.27  222.27         222.27
                2016-07-25  230.01  230.01  230.01  230.01  230.01         230.01
                2016-07-26  225.93  225.93  225.93  225.93  225.93         225.93
            """
            close_panel_pd_cp[col] = benchmark.kl_pd.close
        """
            将复刻后的close_panel_pd_cp与原始close_panel_pd求差后，再进行scaler_std
            ABuScalerUtil.scaler_std(close_panel_pd_cp - close_panel_pd):

                  usFB  usGOOG  usNOAH  usVIPS  usWUBA  us_NYSE:.IXIC
            2015-07-24  0.9705  1.7793  0.7405 -1.6987 -1.9294        -1.0803
            2015-07-27  1.2277  1.6619  1.1473 -1.6270 -1.5697        -0.8853
            2015-07-28  1.1393  1.6826  0.8987 -1.6831 -1.1334        -1.0866
            2015-07-29  0.9629  1.5955  0.7550 -1.7035 -1.0656        -1.2124
            2015-07-30  1.1519  1.5906  0.9265 -1.5197 -1.1169        -1.2878
            ...            ...     ...     ...     ...     ...            ...
            2016-07-21 -1.5539 -0.8188 -0.0710  0.3755  0.5784        -1.2418
            2016-07-22 -1.5899 -0.9012 -0.0644  0.4354  0.5879        -1.3728
            2016-07-25 -1.6371 -0.8138 -0.0746  0.4819  0.4997        -1.3179
            2016-07-26 -1.6473 -0.8509  0.2018  0.3922  0.4085        -1.2702
        """
        regular_diff = ABuScalerUtil.scaler_std(close_panel_pd_cp - close_panel_pd)
        regular_diff.plot()
        plt.title('regular diff')
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.show()

        """
            distance_votes = regular_diff.sum(axis=1):

            投票机制，获取投票coint的差值
                distance_votes
            distance_votes
            2015-07-24   -1.2181
            2015-07-27   -0.0451
            2015-07-28   -0.1825
            2015-07-29   -0.6682
            2015-07-30   -0.2555
                           ...
            2016-07-20   -2.5541
            2016-07-21   -2.7316
            2016-07-22   -2.9049
            2016-07-25   -2.8618
            2016-07-26   -2.7658
            ......................
        """
        distance_votes = regular_diff.sum(axis=1)

        votes_std = distance_votes.std()
        votes_mean = distance_votes.mean()
        above = votes_mean + votes_std
        below = votes_mean - votes_std
        close_regular = ABuScalerUtil.scaler_std(benchmark.kl_pd.close)
        close_regular = (close_regular * distance_votes.max() / 2)
        # noinspection PyUnresolvedReferences
        close_regular.plot()
        distance_votes.plot()

        plt.axhline(votes_mean, color='r')
        plt.axhline(above, color='c')
        plt.axhline(below, color='g')

        plt.title('coint distance votes')
        plt.legend(['close regular', 'distance votes', 'votes mean', 'dvotes above', 'dvotes below'],
                   bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.show()
    return p_value_sorted, sum_rank
